package com.dzeio.charts.views

import android.content.Context
import android.graphics.Canvas
import android.graphics.Paint
import android.graphics.Rect
import android.graphics.RectF
import android.util.AttributeSet
import android.util.Log
import com.dzeio.charts.Animation
import com.dzeio.charts.Entry
import com.dzeio.charts.axis.XAxis
import com.dzeio.charts.axis.YAxis
import kotlin.math.abs
import kotlin.math.max
import kotlin.math.min

class BarChartView @JvmOverloads constructor(context: Context?, attrs: AttributeSet? = null) :
    BaseChart(context, attrs) {

    companion object {
        const val TAG = "DzeioCharts/BarView"
    }

    /**
     * Number of entries displayed at the same time
     */
    private var zoom = 100f

    /**
     * Number of labels displayed at the same time
     */
    var numberOfLabels = 3

    /**
     * Spacing between entries
     */
    var spacing = 22

    val xAxis: XAxis<Double> = XAxis()

    val yAxis: YAxis<Float> = YAxis()

    val animation = Animation()

    private val textTopMargin = 5

    private var barWidth: Int = 0

    private val percentList: ArrayList<Float> = ArrayList()

    /**
     * value goes from 1 to 0 (1 at bottom, 0 at top)
     */
    private var targetPercentList: ArrayList<Float> = ArrayList()

    private var fgPaint: Paint = Paint().also {
        it.isAntiAlias = true
        it.color = yAxis.color
    }
    private val rect: RectF = RectF()

    private var bottomTextDescent = 0
    private var bottomTextHeight = 0

    private var movementOffset: Int = 0

    private val animator: Runnable = object : Runnable {
        override fun run() {
            var needNewFrame = false
            for (i in targetPercentList.indices) {
                val value = animation.updateValue(1f, targetPercentList[i], percentList[i])

                if (value != percentList[i]) {
                    needNewFrame = true
                    percentList[i] = value
                }
            }
            if (needNewFrame) {
                postDelayed(this, animation.getDelay().toLong())
            }
            invalidate()
        }
    }

    var list: ArrayList<Entry> = arrayListOf()

    private var bottomTexts: ArrayList<String> = arrayListOf()

//    init {
//        val mockList = ArrayList<Entry>()
//        for (i in 0 until 25) {
//            mockList.add(Entry(i.toDouble(), i.toFloat()))
//        }
//
//        list = mockList
//
//        this.refresh()
//    }

    fun refresh() {

        val r = Rect()

        //// prepare bottom texts
        bottomTextDescent = 0
        bottomTextHeight = 0
        bottomTexts = arrayListOf()

        //// prepare values

        // set the bar Width (also handle div by 0)
        barWidth = measuredWidth / max(min(list.size, getDisplayedEntries()), 1) - spacing

        // calculate max depending on the maximum value displayed or set in the yAxis params
        val max: Float = if (yAxis.max != null) yAxis.max!! else {
            var calculatedMax = 0f
            for (entry in list.subList(this.getXOffset(), getDisplayedEntries() + this.getXOffset())) {
                if (entry.y > calculatedMax) calculatedMax = entry.y
            }
            calculatedMax
        }

        // make sure the target list
//        Log.d(TAG, list.size.toString())
        targetPercentList = arrayListOf()


        for ((i, item) in list.withIndex()) {
            //// Process bottom texts
            val text = xAxis.onValueFormat(item.x)
            bottomTexts.add(text)

            // get Text boundaries
            xAxis.labels.build().getTextBounds(text, 0, text.length, r)

            // get height of text
            if (bottomTextHeight < r.height()) {
                bottomTextHeight = r.height()
            }

            // get text descent
            val descent = abs(r.bottom)
            if (bottomTextDescent < descent) {
                bottomTextDescent = descent
            }

            //// process values

            // add to animations the values
            targetPercentList.add(min(1 - item.y / max, 1f))
        }

        // post list
        if (percentList.isEmpty() || percentList.size < targetPercentList.size) {
            val temp = targetPercentList.size - percentList.size
            for (i in 0 until temp) {
                percentList.add(1f)
            }
        } else if (percentList.size > targetPercentList.size) {
            val temp = percentList.size - targetPercentList.size
            for (i in 0 until temp) {
                percentList.removeAt(percentList.size - 1)
            }
        }

        // Misc operations

        fgPaint = Paint().apply {
            isAntiAlias = true
            color = yAxis.color
        }

        removeCallbacks(animator)
        post(animator)
    }

    override fun onDraw(canvas: Canvas) {
        if (percentList.isNotEmpty()) {
            // draw each rectangles
            for (i in 1..getDisplayedEntries()) {
//                Log.d(TAG, percentList[i - 1].toString())
                val left = spacing * i + barWidth * (i - 1).toFloat()
//                Log.d(TAG, "$spacing, $i, $barWidth = $left")
                val right = (spacing + barWidth) * i.toFloat()
                val bottom = height - bottomTextHeight - textTopMargin.toFloat()
                val top = bottom * percentList[this.getXOffset() + i - 1]

                // create rounded rect
                canvas.drawRoundRect(left, top, right, bottom, 8f, 8f, fgPaint)
                // remove the bottom corners DUH
                canvas.drawRect(left, max(top, bottom - 8f), right, bottom, fgPaint)
            }
        }
        if (bottomTexts.isNotEmpty() && numberOfLabels > 0) {
            val size = bottomTexts.size
            var i = 1
            var items = size / max(2, (numberOfLabels - 2))

            // handle cases where size is even and numberOfLabels is 3
            if (size % 2 != 0) {
                items += 1
            }

            val rect = Rect()

//            Log.i(TAG, "$size / max($numberOfLabels - 2, 2) = $items")
            for (s in bottomTexts) {

                if ((numberOfLabels <= 2 || i % items != 0) && i != 1 && i != size) {
                    i++
                    continue
                }

//                Log.i(TAG, "Drawing $i")

                xAxis.labels.build().getTextBounds(s, 0, s.length, rect)

                canvas.drawText(
                    s,
                    // handle last entry overflowing
                    min(
                        // handle first entry overflowing
                        max(
                            (spacing * i + barWidth * (i - 1) + barWidth / 2).toFloat(),
                            rect.width() / 2f
                        ),
                        measuredWidth - rect.width() / 2f
                    ),

                    (height - bottomTextDescent).toFloat(),
                    xAxis.labels.build()
                )
                i++
                if (numberOfLabels == 1) {
                    break
                }
            }
        }
    }

    override fun onChartMoved(movementX: Float, movementY: Float) {
        movementOffset = (movementX / 100).toInt()
        refresh()
    }

    override fun onZoomChanged(scale: Float) {
        Log.d(TAG, "New Zoom: $scale")
        zoom = scale
        refresh()
    }

    private fun getXOffset(): Int {
        return min(max(0, xAxis.baseOffset + movementOffset), list.size - 1 - getDisplayedEntries())
    }

    private fun getDisplayedEntries(): Int {
//        Log.d(TAG, "Number of entries displayed ${list.size}, ${xAxis.entriesDisplayed} + (($zoom - 100) * 10) = ${xAxis.entriesDisplayed + ((zoom - 100) * 10).toInt()}")
        return max(1, min(list.size, xAxis.entriesDisplayed + ((zoom - 100) * 10).toInt()))
    }

}
