package com.dzeio.charts.views

import android.content.Context
import android.util.AttributeSet
import android.view.MotionEvent
import android.view.MotionEvent.INVALID_POINTER_ID
import android.view.ScaleGestureDetector
import android.view.View
import androidx.core.view.MotionEventCompat

abstract class BaseChart @JvmOverloads constructor(context: Context?, attrs: AttributeSet? = null) :
    View(context, attrs) {

    // The ‘active pointer’ is the one currently moving our object.
    private var activePointerId = INVALID_POINTER_ID

    private var lastTouchX: Float = 0f
    private var lastTouchY: Float = 0f

    private var posX: Float = 0f
    private var posY: Float = 0f

    private var lastZoom: Float = 100f
    private var currentZoom: Float = 0f

    open fun onChartMoved(movementX: Float, movementY: Float) {}

    /**
     * @param scale Float starting from 100%
     *
     * 99-%  zoom out,
     * 101+% zoom in
     */
    open fun onZoomChanged(scale: Float) {}

    private val scaleGestureDetector = ScaleGestureDetector(
        context,
        object : ScaleGestureDetector.SimpleOnScaleGestureListener() {
            override fun onScale(detector: ScaleGestureDetector): Boolean {

                if (currentZoom != detector.scaleFactor) {
                    currentZoom = detector.scaleFactor
                    onZoomChanged(lastZoom + -currentZoom + 1)
                }

                return super.onScale(detector)
            }

            override fun onScaleEnd(detector: ScaleGestureDetector?) {
                super.onScaleEnd(detector)

                lastZoom += -currentZoom + 1
            }
        })

    /**
     * Code mostly stolen from https://developer.android.com/training/gestures/scale#drag
     */
    override fun onTouchEvent(ev: MotionEvent): Boolean {
        super.onTouchEvent(ev)

        scaleGestureDetector.onTouchEvent(ev)

        when (ev.actionMasked) {
            MotionEvent.ACTION_DOWN -> {
                ev.actionIndex.also { pointerIndex ->
                    // Remember where we started (for dragging)
                    lastTouchX = ev.getX(pointerIndex)
                    lastTouchY = ev.getY(pointerIndex)
                }

                // Save the ID of this pointer (for dragging)
                activePointerId = ev.getPointerId(0)
            }

            MotionEvent.ACTION_MOVE -> {
                // Find the index of the active pointer and fetch its position
                val (x: Float, y: Float) =
                    ev.findPointerIndex(activePointerId).let { pointerIndex ->
                        // Calculate the distance moved
                        ev.getX(pointerIndex) to
                                ev.getY(pointerIndex)
                    }

                posX += x - lastTouchX
                posY += y - lastTouchY

                onChartMoved(-posX, posY)

                // Remember this touch position for the next move event
                lastTouchX = x
                lastTouchY = y
            }
            MotionEvent.ACTION_UP, MotionEvent.ACTION_CANCEL -> {
                activePointerId = INVALID_POINTER_ID
            }
            MotionEvent.ACTION_POINTER_UP -> {

                ev.actionIndex.also { pointerIndex ->
                    ev.getPointerId(pointerIndex)
                        .takeIf { it == activePointerId }
                        ?.run {
                            // This was our active pointer going up. Choose a new
                            // active pointer and adjust accordingly.
                            val newPointerIndex = if (pointerIndex == 0) 1 else 0
                            lastTouchX = MotionEventCompat.getX(ev, newPointerIndex)
                            lastTouchY = MotionEventCompat.getY(ev, newPointerIndex)
                            activePointerId = MotionEventCompat.getPointerId(ev, newPointerIndex)
                        }
                }
            }
        }
        return true
    }
}